/*------------------------------------------------------------------------------
* ppp_ar.c : ppp ambiguity resolution
*
*          Copyright (C) 2012 by T.TAKASU, All rights reserved.
*
* version : $Revision:$ $Date:$
* history : 2012/03/01 1.0  new
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id:$";

/* constants/macros ----------------------------------------------------------*/

#define FIX_FACT        1.5         /* fix factor */
#define MIN_ARC_GAP     300.0       /* min arc gap (s) */
#define MIN_FIX_CNT     20          /* min fix count to fix ambiguity */
#define CONST_AMB       0.02        /* constraint to fixed ambiguity */

#define SQR(x)          ((x)*(x))
#define ROUND(x)        (int)floor((x)+0.5)

#define NP(opt)     ((opt)->dynamics?9:3) /* number of pos solution */
#define IC(s,opt)   (NP(opt)+(s))      /* state index of clocks (s=0:gps,1:glo) */
#define IT(opt)     (IC(0,opt)+NSYS)   /* state index of tropos */
#define NR(opt)     (IT(opt)+((opt)->tropopt<TROPOPT_EST?0:((opt)->tropopt==TROPOPT_EST?1:3)))
                                       /* number of solutions */
#define IB(s,opt)   (NR(opt)+(s)-1)    /* state index of phase bias */

/* wave length of LC (m) -----------------------------------------------------*/
static double lam_LC(int i, int j, int k)
{
    const double f1=FREQ1,f2=FREQ2,f5=FREQ5;
    
    return CLIGHT/(i*f1+j*f2+k*f5);
}
/* carrier-phase LC (m) ------------------------------------------------------*/
static double L_LC(int i, int j, int k, const double *L)
{
    const double f1=FREQ1,f2=FREQ2,f5=FREQ5;
    double L1,L2,L5;
    
    if ((i&&!L[0])||(j&&!L[1])||(k&&!L[2])) {
        return 0.0;
    }
    L1=CLIGHT/f1*L[0];
    L2=CLIGHT/f2*L[1];
    L5=CLIGHT/f5*L[2];
    return (i*f1*L1+j*f2*L2+k*f5*L5)/(i*f1+j*f2+k*f5);
}
/* pseudorange LC (m) --------------------------------------------------------*/
static double P_LC(int i, int j, int k, const double *P)
{
    const double f1=FREQ1,f2=FREQ2,f5=FREQ5;
    double P1,P2,P5;
    
    if ((i&&!P[0])||(j&&!P[1])||(k&&!P[2])) {
        return 0.0;
    }
    P1=P[0];
    P2=P[1];
    P5=P[2];
    return (i*f1*P1+j*f2*P2+k*f5*P5)/(i*f1+j*f2+k*f5);
}
/* noise variance of LC (m) --------------------------------------------------*/
static double var_LC(int i, int j, int k, double sig)
{
    const double f1=FREQ1,f2=FREQ2,f5=FREQ5;
    
    return (SQR(i*f1)+SQR(j*f2)+SQR(k*f5))/SQR(i*f1+j*f2+k*f5)*SQR(sig);
}
/* zero-difference noise variance --------------------------------------------*/
static double ZD_var(double var, double el)
{
    double sinel=sin(el);
    return var+var/sinel/sinel;
}
/* average LC ----------------------------------------------------------------*/
static void average_LC(rtk_t *rtk, const obsd_t *obs, int n, const nav_t *nav,
                       const double *azel)
{
    ambc_t *amb;
    double LC1,LC2,LC3,var1,var2,var3,err=rtk->opt.err[1]*100.0;
    int i,sat;
    
    for (i=0;i<n;i++) {
        sat=obs[i].sat;
        
#if 0
        if (azel[1+2*i]<rtk->opt.elmin) continue;
#else
        if (azel[1+2*i]<5.0*D2R) continue;
#endif
        if (satsys(sat,NULL)!=SYS_GPS) continue;
        
        /* triple-freq carrier and code LC (m) */
        LC1=L_LC(1,-1, 0,obs[i].L)-P_LC(1,1,0,obs[i].P);
        LC2=L_LC(0, 1,-1,obs[i].L)-P_LC(0,1,1,obs[i].P);
        LC3=L_LC(1,-6, 5,obs[i].L)-P_LC(1,1,0,obs[i].P);
        
        /* measurement noise variance (m) */
        var1=ZD_var(var_LC(1,1,0,err),azel[1+2*i]);
        var2=ZD_var(var_LC(0,1,1,err),azel[1+2*i]);
        var3=ZD_var(var_LC(1,1,0,err),azel[1+2*i]);
        
        amb=rtk->ambc+sat-1;
        
        if (rtk->ssat[sat-1].slip[0]||rtk->ssat[sat-1].slip[1]||
            rtk->ssat[sat-1].slip[2]||amb->n[0]==0||
            fabs(timediff(amb->epoch[0],obs[0].time))>MIN_ARC_GAP) {
            
            amb->fixcnt=0;
            if (LC1) {
                amb->n[0]=1; amb->LC[0]=LC1; amb->LCv[0]=var1;
            }
            if (LC2) {
                amb->n[1]=1; amb->LC[1]=LC2; amb->LCv[1]=var2;
            }
            if (LC3) {
                amb->n[2]=1; amb->LC[2]=LC3; amb->LCv[2]=var3;
            }
        }
        else { /* averaging */
            if (LC1) {
                amb->LC [0]+=(LC1-amb->LC[0])/(++amb->n[0]);
                amb->LCv[0]+=(var1-amb->LCv[0])/amb->n[0];
            }
            if (LC2) {
                amb->LC [1]+=(LC2-amb->LC[1])/(++amb->n[1]);
                amb->LCv[1]+=(var2-amb->LCv[1])/amb->n[1];
            }
            if (LC3) {
                amb->LC [2]+=(LC3-amb->LC[2])/(++amb->n[2]);
                amb->LCv[2]+=(var3-amb->LCv[2])/amb->n[2];
            }
        }
        amb->epoch[0]=obs[0].time;
    }
}
/* fix solution --------------------------------------------------------------*/
static int fixsol(rtk_t *rtk, const double *v, const double *H, int nv)
{
    double *R;
    int i,j,info;
    
    if (nv<=0) return 0;
    
    R=zeros(nv,nv);
    
    for (i=0;i<nv;i++) R[i+i*nv]=SQR(CONST_AMB);
    
    /* update states with constraints */
    if (!(info=filter(rtk->x,rtk->P,H,v,R,rtk->nx,nv))) {
        
        /* set solution */
        for (i=0;i<rtk->na;i++) {
            rtk->xa[i]=rtk->x[i];
            for (j=0;j<rtk->na;j++) {
                rtk->Pa[i+j*rtk->na]=rtk->Pa[j+i*rtk->na]=rtk->P[i+j*rtk->nx];
            }
        }
    }
    else {
        trace(1,"filter error (info=%d)\n",info);
        nv=0;
    }
    free(R);
    return nv;
}
/* linear dependency check ---------------------------------------------------*/
static int is_depend(int sat1, int sat2, int *flgs, int *max_flg)
{
    int i;
    
    if (flgs[sat1-1]==0&&flgs[sat2-1]==0) {
        flgs[sat1-1]=flgs[sat2-1]=++(*max_flg);
    }
    else if (flgs[sat1-1]==0&&flgs[sat2-1]!=0) {
        flgs[sat1-1]=flgs[sat2-1];
    }
    else if (flgs[sat1-1]!=0&&flgs[sat2-1]==0) {
        flgs[sat2-1]=flgs[sat1-1];
    }
    else if (flgs[sat1-1]>flgs[sat2-1]) {
        for (i=0;i<MAXSAT;i++) if (flgs[i]==flgs[sat2-1]) flgs[i]=flgs[sat1-1];
    }
    else if (flgs[sat1-1]<flgs[sat2-1]) {
        for (i=0;i<MAXSAT;i++) if (flgs[i]==flgs[sat1-1]) flgs[i]=flgs[sat2-1];
    }
    else return 0; /* linear depenent */
    return 1;
}
/* resolve integer ambiguity by WL-NL ----------------------------------------*/
extern int pppamb(rtk_t *rtk, const obsd_t *obs, int n, const nav_t *nav,
                  const double *azel)
{
    ambc_t *ambi,*ambj;
    double lam_N,lam_W,C1,C2,BW,vW,BC,vC,B1,v1,*H,*v;
    int i,j,k,l,NW,N1,N2,nv=0,fix;
    
    if (n<=0||rtk->opt.ionoopt!=IONOOPT_IFLC||rtk->opt.nf<2) return 0;
    
    trace(3,"resamb_WLNL: time=%s n=%d\n",time_str(obs[0].time,0),n);
    
    lam_N=lam_LC(1, 1,0);
    lam_W=lam_LC(1,-1,0);
    C1= SQR(lam[1])/(SQR(lam[1])-SQR(lam[0]));
    C2=-SQR(lam[0])/(SQR(lam[1])-SQR(lam[0]));
    
    v=zeros(n*n,1); H=zeros(rtk->nx,n*n);
    
    /* average LC */
    average_LC(rtk,obs,n,nav,azel);
    
    for (i=0;i<n-1;i++) for (j=i+1;j<n;j++) {
        if (azel[1+j*2]<rtk->opt.elmin||
           (rtk->opt.elmaskar>0.0&&azel[1+j*2]<rtk->opt.elmaskar)) continue;
        ambi=rtk->ambc+obs[i].sat-1;
        ambj=rtk->ambc+obs[j].sat-1;
        k=IB(obs[i].sat,&rtk->opt);
        l=IB(obs[j].sat,&rtk->opt);
        
        if (!ambi->n[0]||!ambj->n[0]) continue;
        
        /* wide lane ambiguity (cycle) */
        BW=(ambi->LC[0]-ambj->LC[0])/lam_W+
           (nav->wlbias[obs[i].sat-1]-nav->wlbias[obs[j].sat-1]);
        vW=(ambi->LCv[0]/ambi->n[0]+ambj->LCv[0]/ambj->n[0])/SQR(lam_W);
        NW=ROUND(BW);
        
        /* narrow lane ambiguity (cycle) */
        BC=rtk->x[k]-rtk->x[l];
        vC=rtk->P[k+k*rtk->nx]+rtk->P[l+l*rtk->nx]-2.0*rtk->P[k+l*rtk->nx];
        B1=(BC+C2*lam[1]*NW)/lam_N;
        v1=vC/SQR(lam_N);
        N1=ROUND(B1); N2=N1-NW;
        
#if 0
        fix=fabs(NW-BW)<=rtk->opt.thresar*FIX_FACT&&sqrt(vW)<=rtk->opt.thresar&&
            fabs(N1-B1)<=rtk->opt.thresar*FIX_FACT&&sqrt(v1)<=rtk->opt.thresar;
#else
        fix=fabs(NW-BW)<=0.25&&sqrt(vW)<=0.25&&
            fabs(N1-B1)<=0.25&&sqrt(v1)<=0.25;
#endif
        
        trace(1,"%s sat=%2d-%2d B=%13.3f %13.3f sig=%7.3f %7.3f n=%4d %4d %s\n",
              time_str(obs[0].time,0),obs[i].sat,obs[j].sat,BW,B1,sqrt(vW),
              sqrt(v1),rtk->ambc[obs[i].sat-1].n[0],rtk->ambc[obs[j].sat-1].n[0],
              fix?"FIX":"---");
        
        ambi->fixcnt++;
        ambj->fixcnt++;
        if (!fix||ambi->fixcnt<MIN_FIX_CNT||ambj->fixcnt<MIN_FIX_CNT) continue;
        
#if 1
        /* constraint to dd-ambiguity */
        v[nv]=(C1*lam[0]*N1+C2*lam[1]*N2)-(rtk->x[k]-rtk->x[l]);
        H[k+nv*rtk->nx]= 1.0;
        H[l+nv*rtk->nx]=-1.0;
        nv++;
#endif
    }
    /* fix ppp solution */
    nv=fixsol(rtk,v,H,nv);
    
    free(v); free(H);
    
    return nv>=n/3; /* fix if one third ambiguities fixed */
}
